import sys
import time
import datetime
import argparse

from kabaret.app.session import KabaretSession
from kabaret.app.actors.flow import Flow
from kabaret import flow


class CleanSyncJobsSession(KabaretSession):

    def _create_actors(self):
        Flow(self)
    
    def clear_jobs(self, project_name, clear_every=900, clear_before=1800):
        clear_action_oid = f'/{project_name}/admin/multisites/working_sites/clear_site_queues'

        try:
            valid_project = self.cmds.Flow.is_action(clear_action_oid)
        except (flow.MissingChildError, flow.MissingRelationError):
            valid_project = False
        
        if not valid_project:
            self.log_error('Invalid project !')
            return

        try:
            while True:
                self.log_info('Clearing jobs (%s)...' % datetime.datetime.fromtimestamp(time.time()))
                self.cmds.Flow.set_value(oid=clear_action_oid+'/emitted_since', value=clear_before)
                self.cmds.Flow.run_action(oid=clear_action_oid, button='Clear')
                time.sleep(clear_every)

        except KeyboardInterrupt:
            self.log_info('Cleaner stopped. Exiting...')
            return


def process_remaining_args(args):
    parser = argparse.ArgumentParser(
        description='Libreflow Session Arguments'
    )
    parser.add_argument(
        '-p', '--clear-period', default=900, dest='period'
    )
    parser.add_argument(
        '-t', '--elapsed-time', default=1800, dest='time'
    )
    values, remaining_args = parser.parse_known_args(args)
    
    return (float(values.period), float(values.time), remaining_args)


def main(argv):
    (
        session_name,
        host,
        port,
        cluster_name,
        db,
        password,
        debug,
        remaining_args,
    ) = CleanSyncJobsSession.parse_command_line_args(argv)

    session = CleanSyncJobsSession(session_name=session_name, debug=debug)
    session.cmds.Cluster.connect(host, port, cluster_name, db, password)

    (
        period,
        time,
        remaining_args
    ) = process_remaining_args(remaining_args)

    if remaining_args:
        project_name = remaining_args[0]
    else:
        project_name = 'siren'

    # Check project existence
    if not session.cmds.Flow.exists('/'+project_name):
        session.log_error(
            f'No project /{project_name} found on this cluster. ' \
            'Please specify an existing project name.'
        )
        return

    # Check project root type
    project_type_name = session.get_actor('Flow').get_project_qualified_type_name(project_name)

    session.clear_jobs(project_name, clear_every=period, clear_before=time)


if __name__ == "__main__":
    main(sys.argv[1:])


import os
import json
import gazu
import getpass
import pathlib
import fnmatch
import re

from kabaret import flow
from kabaret.app.ui.gui.widgets.flow.flow_view import (
    CustomPageWidget,
    QtWidgets,
    QtCore,
)
from kabaret.app.ui.gui.widgets.editors import editor_factory
from kabaret.flow_contextual_dict import ContextualView, get_contextual_dict
from kabaret.subprocess_manager.runners import Explorer
from kabaret.flow_entities.store import EntityStore

from .users import Users, UserEnvironment, UserBookmarks, UserProfile
from .film import Sequences
from .lib import Assets
from .runners import (
    Krita,
    Blender,
    AfterEffects,
    AfterEffectsRender,
    VSCodium,
    NotepadPP,
    MarkSequenceRunner,
    Firefox,
    SessionWorker,
    DefaultEditor,
    ImageMagick,
    RV,
    DefaultRunners,
)
from .site import (
    Synchronize, GotoCurrentSiteQueue, ShowSiteSyncErrors,
    WorkingSites, ExchangeSites,
    MinioFileUploader, MinioFileDownloader,
    RequestRevisions, MultiRequestRevisions,
)
from .file import TrackedFile, Revision, FileManager, DefaultFileMap
from .kitsu import KitsuAPIWrapper, KitsuBindings
from .dependency import DependencyTemplates


class LoginPageWidget(CustomPageWidget):
    def build(self):
        # Get project root oid
        self.project_oid = self.session.cmds.Flow.split_oid(self.oid)[0][1]

        # Build UI
        self.label = QtWidgets.QLabel(self)
        self.label.setText("<h2>Connexion page</h2>")
        self.error_label = QtWidgets.QLabel(self)
        self.error_label.setText("")

        self.lineedit_kitsu_login = QtWidgets.QLineEdit()
        self.lineedit_kitsu_password = QtWidgets.QLineEdit()
        self.button_login = QtWidgets.QPushButton("Log in")
        self.button_login.setMinimumWidth(100)

        # Set kitsu_password field input mode
        self.lineedit_kitsu_password.setEchoMode(QtWidgets.QLineEdit.Password)
        self.lineedit_kitsu_password.setInputMethodHints(
            QtCore.Qt.ImhHiddenText
            | QtCore.Qt.ImhNoPredictiveText
            | QtCore.Qt.ImhNoAutoUppercase
        )

        grid_layout = QtWidgets.QGridLayout()
        grid_layout.addWidget(self.label, 0, 0)
        grid_layout.addWidget(self.lineedit_kitsu_login, 1, 0, 1, 2)
        grid_layout.addWidget(self.lineedit_kitsu_password, 2, 0, 1, 2)
        grid_layout.addWidget(self.error_label, 3, 0)
        grid_layout.addWidget(self.button_login, 3, 1)
        grid_layout.setColumnStretch(0, 1)

        self.setLayout(grid_layout)

        self.button_login.clicked.connect(self.on_connection_validated)
        self.lineedit_kitsu_login.returnPressed.connect(self.on_connection_validated)
        self.lineedit_kitsu_password.returnPressed.connect(self.on_connection_validated)

    def on_connection_validated(self):
        kitsu_login = self.lineedit_kitsu_login.text()
        kitsu_password = self.lineedit_kitsu_password.text()
        logged_in = self.session.cmds.Flow.call(
            self.project_oid,
            "log_in",
            args={},
            kwargs={"kitsu_login": kitsu_login, "kitsu_password": kitsu_password},
        )

        if logged_in:
            if logged_in == 1:
                text = "<font color=#C71B1F>Login failed... Check credentials !</font>"
            elif logged_in == 2:
                support = self.session.cmds.Flow.get_value(self.project_oid + "/admin/project_settings/support")
                text = "<font color=#D66700>A user with the same name but a different Kitsu ID is already registered for this project."
                if support:
                    text += f"<br>We invite you to contact the support of the project ({support}).</font>"
                else:
                    text += "</font>"
            
            self.error_label.setText(text)
            return

        self.error_label.setText("")
        self.page.refresh()


class LoginPage(flow.Object):

    kitsu_server_url = flow.Param("")



class SynchronizeFilesResult(flow.Action):

    def allow_context(self, context):
        return False

    def get_buttons(self):
        return ["Close"]
    
    def run(self, button):
        return


class SynchronizeFiles(Synchronize):
    
    ICON = ('icons.libreflow', 'sync')

    result = flow.Child(SynchronizeFilesResult)

    def needs_dialog(self):
        return True
    
    def get_buttons(self):
        exchange_site = self.root().project().get_exchange_site()
        
        if not exchange_site.configured.get():
            self.message.set((
                "<h3><font color=#D5000D>"
                "Exchange site not configured"
                "</font></h3>"
                "Please configure the exchange site of this project."
            ))
            return ['Configure exchange site', 'Cancel']
        
        self.message.set((
            "<h3><font color=#D66700>"
            "Synchronizing all requested files may freeze your session for a while."
            "</font></h3>"
        ))
        return ["Confirm", "Cancel"]

    def allow_context(self, context):
        return context and context.endswith(".details")
    
    def run(self, button):
        if button == "Cancel":
            return
        elif button == 'Configure exchange site':
            exchange_site = self.root().project().get_exchange_site()
            return self.get_result(next_action=exchange_site.configuration.oid())

        # Save initial job counts
        init_job_count = self._compute_job_count()

        # Process jobs
        super(SynchronizeFiles, self).run(button)

        # Print final counts
        nb_waiting_init = init_job_count['WAITING']
        nb_processed = self._job_count['PROCESSED'] - init_job_count['PROCESSED']
        nb_error_init = init_job_count['ERROR']
        nb_error = self._job_count['ERROR'] - nb_error_init

        if nb_processed == nb_waiting_init and nb_error == 0:
            msg = (
                "<h3><font color=#029600>"
                "Synchronization successful !"
                "</font></h3>"
            )
        else:
            msg = ''
            
            if nb_processed < nb_waiting_init:
                msg += (
                    '<h3><font color=#D66700>'
                    f'Synchronized files: {nb_processed}/{nb_waiting_init}'
                    '</font></h3>'
                )
            if nb_error > 0:
                msg += (
                    '<h3><font color=#D5000D>'
                    f'Errors: {nb_error}/{nb_waiting_init}'
                    '</font></h3>'
                )
        
        if nb_error_init > 0:
            msg += (
                '<h3><font color=#D5000D>'
                f'Other waiting errors: {nb_error_init}'
                '</font></h3>'
            )
            
        self.result.message.set(msg)
        
        return self.get_result(next_action=self.result.oid())


class GotoBookmarks(flow.Action):
    ICON = ("icons.gui", "star")

    def needs_dialog(self):
        return False
    
    def run(self, button):
        return self.get_result(
            goto=self.root().project().admin.bookmarks.oid()
        )


class WorkingSiteChoiceValue(flow.values.ChoiceValue):

    _parent = flow.Parent()

    def choices(self):
        return self.root().project().get_working_sites().mapped_names()


class CleanZipFiles(flow.Action):

    ICON = ('icons.libreflow', 'clean')

    def __init__(self, parent, name):
        super(CleanZipFiles, self).__init__(parent, name)
        self._paths = []

    def get_buttons(self):
        self.message.set('<h2>Remove ZIP files</h2>')
        return ['Confirm', 'Search', 'Cancel']
    
    def run(self, button):
        if button == 'Cancel':
            return
        elif button == 'Search':
            root_project = self.root().project().get_root().replace('\\', '/')
            self._paths = []
            file_count = 0
            file_size = 0
            
            root_lib = os.path.join(root_project, 'lib').replace('\\', '/')
            root_film = os.path.join(root_project, 'siren').replace('\\', '/')
            
            for root_path in (root_lib, root_film):
                for root, _, files in os.walk(root_path):
                    norm_root = root.replace('\\', '/').replace(root_project, '')
                    
                    for f in files:
                        match = re.search('lib/(chars|props|sets)/.*/.*/.*/.*/v\d\d\d', norm_root)
                        match = match or re.search('/siren/sq\d\d/sc\d\d\d\d/.*/.*/v\d\d\d', norm_root)
                        
                        if match is not None and os.path.splitext(f)[1] == '.zip':
                            path = os.path.join(root, f)
                            file_count += 1
                            file_size += os.path.getsize(path)
                            self._paths.append(path)
            
            self.message.set((
                '<h2>Remove ZIP files</h2>'
                '<b>%i</b> ZIP files found in this project (~%03f Gb).'
                '<h3><font color="#D5000D">Remove all ?</font></h3>' % (file_count, file_size * 0.000000001)
            ))

            return self.get_result(close=False)
        else:
            for p in self._paths:
                try:
                    self.root().session().log_warning('Remove file %s' % p)
                    os.remove(p)
                except Exception:
                    self.root().session().log_warning('Failed to remove file %s' % p)


class Synchronization(flow.Object):

    ICON = ("icons.libreflow", "sync")
    
    _project = flow.Parent()
    synchronize_files = flow.Child(SynchronizeFiles)
    show_jobs = flow.Child(GotoCurrentSiteQueue)
    show_errors = flow.Child(ShowSiteSyncErrors)
    request_revisions = flow.Child(RequestRevisions)
    multi_request_revisions = flow.Child(MultiRequestRevisions)

    def __init__(self, parent, name):
        super(Synchronization, self).__init__(parent, name)
        self._initialized = False
    
    def _ensure_waiting_count(self):
        jobs = self._project.get_current_site().queue.job_list

        if not self._initialized:
            jobs.update_waiting_count()
            self._initialized = True
        
        return jobs.count_waiting.get()

    def summary(self):
        nb_waiting_jobs = self.root().project().get_current_site().count_jobs(
            status='WAITING'
        )
        
        if nb_waiting_jobs > 0:
            return (
                "<font color=#D5000D><b>"
                f"{nb_waiting_jobs} job(s) waiting"
                "</b></font>"
            )


class LogOut(flow.Action):

    ICON = ("icons.libreflow", "log_out")

    def needs_dialog(self):
        return False
    
    def allow_context(self, context):
        return context and context.endswith(".details")
    
    def run(self, button):
        project = self.root().project()
        kitsu_config = "%s/kitsu_config.json" % project.user_settings_folder()

        if os.path.exists(kitsu_config):
            os.remove(kitsu_config)

        project.log_out()
        
        return self.get_result(
            goto=project.oid(),
            refresh=True
        )


class KitsuConfig(flow.Object):

    ICON = ('icons.libreflow', 'kitsu')

    server_url = flow.Param().watched()
    project_name = flow.Param().watched()
    project_id = flow.Computed(
        cached=True
    )
    uploadable_files = flow.Param("")
    bindings = flow.Child(KitsuBindings)
    gazu_api = flow.Child(KitsuAPIWrapper).ui(
        hidden=True
    )
    configured = flow.Computed().ui(
        editor='bool',
        hidden=True
    )

    def child_value_changed(self, child_value):
        self.project_id.touch()

    def compute_child_value(self, child_value):
        if child_value is self.project_id:
            self.project_id.set(
                self.gazu_api.get_project_id()
            )
        elif child_value is self.configured:
            self.configured.set(
                self.project_id.get() is not None
            )
    
    def is_uploadable(self, file_name):
        patterns = self.uploadable_files.get().split(',')

        for pattern in patterns:
            pattern = pattern.replace(' ', '')
            if not fnmatch.fnmatch(file_name, pattern):
                return False
    
    def summary(self):
        if not self.configured.get():
            return (
                '<font color=#D5000D>'
                '<b>Server URL</b> and <b>project name</b> '
                'unset: Kitsu features won\'t be available'
            )


class ProjectSettings(flow.Object):

    ICON = ('icons.gui', 'settings')

    project_nice_name = flow.Param()
    project_thumbnail = flow.Param().ui(editor='textarea')
    support = flow.Param("")

    settings = flow.Child(ContextualView)
    default_files = flow.Child(DefaultFileMap)
    non_editable_files = flow.Param("")
    auto_upload = flow.Param('').ui(
        tooltip='Enable <b>upload after publish</b> by default for all files matching the provided patterns')
    hide_upload_after_publish = flow.Param('')
    clean_zip_files = flow.Child(CleanZipFiles).ui(label='Clean ZIP files')
    enable_file_lock = flow.BoolParam(True)

    def get_auto_upload_files(self):
        return self.auto_upload.get().replace(' ', '').split(',')
    
    def get_hidden_upload_files(self):
        return self.hide_upload_after_publish.get().replace(' ', '').split(',')


class ExchangeSiteChoiceValue(flow.values.ChoiceValue):

    def choices(self):
        return self.root().project().get_exchange_sites().mapped_names()


class MultisiteConfig(flow.Object):

    ICON = ('icons.gui', 'sitemap')

    current_site_name = flow.Computed(cached=True)
    root_dir = flow.Computed(cached=True)
    temp_dir = flow.Computed(cached=True)
    exchange_site_name = flow.Param("default_exchange", ExchangeSiteChoiceValue).watched().ui(
        editable=False
    )
    working_sites = flow.Child(WorkingSites).ui(label="Working sites")
    exchange_sites = flow.Child(ExchangeSites).ui(label="Exchange sites")
    sites_data = flow.Computed(cached=True).ui(hidden=True)

    def child_value_changed(self, child_value):
        if child_value is self.exchange_site_name:
            self.touch()

    def compute_child_value(self, child_value):
        if child_value is self.root_dir:
            '''
            TODO : Override by site on multisite !
            '''
            root_dir = None
            if "ROOT_DIR" in os.environ:
                print("WARNING: ROOT_DIR was defined by the environement !")
                root_dir = os.environ["ROOT_DIR"]
            else:
                # Otherwise, get current site's root dir
                root_dir = self.working_sites[self.current_site_name.get()].root_folder.get()
            
            child_value.set(root_dir)
        elif child_value is self.temp_dir:
            path = self.root_dir.get()

            if path is not None:
                path = os.path.join(path, '.tmp')

                if not os.path.exists(path):
                    os.makedirs(path)
            
            self.temp_dir.set(path)
        elif child_value is self.current_site_name:
            if "KABARET_SITE_NAME" in os.environ:
                site_name = os.environ["KABARET_SITE_NAME"]
            else:
                site_name = os.environ["KABARET_CLUSTER_NAME"]

            if not self.working_sites.has_mapped_name(site_name):
                self.root().session().log_warning((
                    f"Site {site_name} not found in project working sites. "
                    "Falling back to default."
                ))
                site_name = "default"

            child_value.set(site_name)
        elif child_value is self.sites_data:
            self.sites_data.set(self._compute_sites_data())
    
    def _compute_sites_data(self):
        data = {}

        for s in self.root().project().get_working_sites().mapped_items():
            data[s.name()] = {
                'is_default': True,
                'is_active': s.is_active.get(),
                'short_name': s.short_name.get()
            }
        for s in self.root().project().get_exchange_sites().mapped_items():
            data[s.name()] = {
                'is_default': True,
                'is_active': s.is_active.get(),
                'short_name': s.short_name.get()
            }
        
        return data

    def summary(self):
        current_site = self.working_sites[self.current_site_name.get()]
        exchange_site = self.exchange_sites[self.exchange_site_name.get()]

        if not current_site.configured.get():
            return (
                '<font color=#D5000D>'
                f'Current site <b>{current_site.name()}</b> '
                'needs to be configured</font>'
            )
        
        if not exchange_site.configured.get():
            return (
                '<font color=#D5000D>'
                f'Exchange site <b>{exchange_site.name()}</b> '
                'needs to be configured.</font>'
            )


class Admin(flow.Object):

    ICON = ("icons.gui", "team-admin")

    _project = flow.Parent()

    project_settings = flow.Child(ProjectSettings).injectable()
    users = flow.Child(Users).ui(show_filter=True)
    multisites = flow.Child(MultisiteConfig).ui(label='Sites')
    kitsu = flow.Child(KitsuConfig)

    dependency_templates = flow.Child(DependencyTemplates)
    
    user_environment = flow.Child(UserEnvironment).ui(
        expanded=False,
        label='Session environment')
    default_applications = flow.Child(DefaultRunners).ui(expanded=False).injectable()

    login_page = flow.Child(LoginPage).ui(hidden=True)

    file_manager = flow.Child(FileManager)
    store = flow.Child(EntityStore)


class Project(flow.Object):

    log_out_action = flow.Child(LogOut).ui(label="Log out")

    user = flow.Child(UserProfile)
    asset_lib = flow.Child(Assets).ui(expanded=True)
    sequences = flow.Child(Sequences).ui(expanded=True)
    admin = flow.Child(Admin)
    synchronization = flow.Child(Synchronization).ui(expanded=True)

    _RUNNERS_FACTORY = None

    def get_root(self, alternative=None):
        '''
        alternative can be used if root_dir.get() returns None
        '''
        root_dir = self.admin.multisites.root_dir.get()
        if root_dir is None and alternative != None:
            root_dir = alternative
        return root_dir
    
    def get_temp_folder(self):
        return self.admin.multisites.temp_dir.get()

    def set_user_name(self, username):
        '''
        Sets the name of the current user.
        '''
        project_settings_folder = self.project_settings_folder()

        if not os.path.exists(project_settings_folder):
            os.makedirs(project_settings_folder)

        user_file = os.path.join(project_settings_folder, "current_user.json")

        with open(user_file, "w+") as f:
            user_config = dict(username=username)
            json.dump(user_config, f)
        
        self.user.current_user_id.touch()

    def get_user_name(self):
        '''
        Returns the name of the current user.
        '''
        return self.user.current_user_id.get()

    def get_user(self, name=None):
        '''
        If provided, returns the user of this project with the
        given `name`. Return the project's current user otherwise.

        In any case, None is returned if no user is found.
        '''
        if name is None:
            name = self.get_user_name()
        
        if not self.admin.users.has_mapped_name(name):
            return None
        
        return self.admin.users[name]

    def get_current_site(self):
        """
        Returns the site within which the current
        session is beeing run.
        """
        return self.get_working_site(
            self.admin.multisites.current_site_name.get()
        )
    
    def get_working_site(self, name):
        """
        Returns the working site of the project
        with the given name.
        """
        return self.get_working_sites()[name]
    
    def get_exchange_site(self):
        """
        Returns the first exchange site found in
        project's registered site list.
        """
        return self.get_exchange_sites()[
            self.admin.multisites.exchange_site_name.get()
        ]

    def get_working_sites(self):
        return self.admin.multisites.working_sites

    def get_exchange_sites(self):
        return self.admin.multisites.exchange_sites
    
    def get_site_names(self, short_names=False):
        working_sites = self.get_working_sites()
        exchange_sites = self.get_exchange_sites()

        if short_names:
            return working_sites.short_names() + exchange_sites.short_names()

        return working_sites.mapped_names() + exchange_sites.mapped_names()

    def is_admin(self, username):
        try:
            return self.admin.users.is_admin(username)
        except flow.exceptions.MappedNameError:
            # Unregistered user not admin by default
            return False

    def show_login_page(self):
        kitsu_api = self.kitsu_api()
        valid_host = kitsu_api.host_is_valid()

        if not valid_host:
            logged_in = False
        else:
            logged_in = kitsu_api.current_user_logged_in()

        if not logged_in:
            try:
                f = open("%s/kitsu_config.json" % self.user_settings_folder(), "r")
            except IOError:
                logged_in = False
            else:
                kitsu_config = json.load(f)
                kitsu_api.set_host(kitsu_config["kitsu_host"])
                kitsu_api.set_tokens(kitsu_config["tokens"])

                logged_in = kitsu_api.current_user_logged_in()

        return not kitsu_api.host_is_valid() or not logged_in

    def update_kitsu_host(self, server_url):
        kitsu_api = self.kitsu_api()
        kitsu_api.set_server_url(server_url)
        kitsu_api.set_host(server_url + "/api")

        return kitsu_api.host_is_valid()

    def log_in(self, kitsu_login, kitsu_password):
        kitsu_api = self.kitsu_api()

        self.update_kitsu_host(self.admin.kitsu.server_url.get())
        kitsu_user_data = kitsu_api.log_in(kitsu_login, kitsu_password)

        if kitsu_user_data is None:
            return 1
        
        kitsu_id = kitsu_user_data['user']['id']

        # Set current user
        # user_name = self.admin.users.get_user_name(kitsu_login)
        user = self.admin.users.get_user_by_uid(kitsu_id)

        if user is None:
            # Create user profile if not registered
            user = self.admin.users.add_user(kitsu_id, kitsu_login)

            if user is None:
                return 2

            # user_name = user.name()
        
        self.set_user_name(user.name())

        # Create user folder at first login
        user_settings_folder = self.user_settings_folder()
        if not os.path.exists(user_settings_folder):
            os.makedirs(user_settings_folder)

        # Save authentification tokens
        tokens = kitsu_api.get_tokens()
        kitsu_config = {}
        kitsu_config["tokens"] = tokens
        kitsu_config["kitsu_host"] = kitsu_api.get_host()

        if user_settings_folder is not None:
            with open("%s/kitsu_config.json" % user_settings_folder, "w+") as f:
                json.dump(kitsu_config, f)

        return 0
    
    def log_out(self):
        self.kitsu_api().log_out()
    
    def kitsu_id(self):
        return self.admin.kitsu.project_id.get()

    def kitsu_name(self):
        return self.admin.kitsu.project_name.get()
    
    def kitsu_config(self):
        return self.admin.kitsu
    
    def kitsu_api(self):
        return self.admin.kitsu.gazu_api

    def kitsu_bindings(self):
        return self.admin.kitsu.bindings


    def get_contextual_view(self, context_name):
        if context_name == "settings":
            return self.admin.project_settings.settings

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(
                PROJECT=self.name(),
                default_shot_layout_files="{sequence}_{shot}_layout.blend, {sequence}_{shot}_layout-movie.mov",
                default_shot_animation_files="{sequence}_{shot}_animation.blend, {sequence}_{shot}_animation-movie.mov, {sequence}_{shot}_animation-export.abc",
                default_asset_model_files="{asset_name}_model.blend, {asset_name}_model-movie.mov, {asset_name}_model-export.abc",
                default_asset_rig_files="{asset_name}_rig.blend, {asset_name}_rig-turnaround.mov",
                project_thumbnail="{ROOT_DIR}TECH/{PROJECT}_thumbnail.png",
            )

    def project_settings_folder(self):
        return os.path.join(
            pathlib.Path.home(),
            ".libreflow",
            self.name()
        )

    def user_settings_folder(self):
        '''
        DEPRECATED ! TO DELETE ? 
        '''
        return os.path.join(
            self.project_settings_folder(),
            self.get_user_name()
        )
    
    def get_default_file_presets(self):
        '''
        Must return an instance of libreflow.baseflow.file.DefaultFileMap.
        '''
        return self.admin.project_settings.default_files

    def get_project_thumbnail2(self):
        image = self.admin.project_settings.project_thumbnail.get()
        return image
    def get_project_thumbnail(self):
        contextual_dict = get_contextual_dict(self, "settings")
        contextual_dict["ROOT_DIR"] = self.get_root()
        if "project_thumbnail" not in contextual_dict:
            return None
        path = None

        try:
            path = contextual_dict["project_thumbnail"].format(**contextual_dict)
        except KeyError:
            return None

        if path and os.path.exists(path):
            return path
        else:
            return None

    def update_user_environment(self):
        self.admin.user_environment.update()

    def _register_runners(self):
        self._RUNNERS_FACTORY.ensure_runner_type(Blender)
        self._RUNNERS_FACTORY.ensure_runner_type(Krita)
        self._RUNNERS_FACTORY.ensure_runner_type(AfterEffects)
        self._RUNNERS_FACTORY.ensure_runner_type(AfterEffectsRender)
        self._RUNNERS_FACTORY.ensure_runner_type(VSCodium)
        self._RUNNERS_FACTORY.ensure_runner_type(NotepadPP)
        self._RUNNERS_FACTORY.ensure_runner_type(MarkSequenceRunner)
        self._RUNNERS_FACTORY.ensure_runner_type(Firefox)
        self._RUNNERS_FACTORY.ensure_runner_type(Explorer)
        self._RUNNERS_FACTORY.ensure_runner_type(MinioFileUploader)
        self._RUNNERS_FACTORY.ensure_runner_type(MinioFileDownloader)
        self._RUNNERS_FACTORY.ensure_runner_type(SessionWorker)
        self._RUNNERS_FACTORY.ensure_runner_type(DefaultEditor)
        self._RUNNERS_FACTORY.ensure_runner_type(ImageMagick)
        self._RUNNERS_FACTORY.ensure_runner_type(RV)

    def ensure_runners_loaded(self):
        session = self.root().session()
        subprocess_manager = session.get_actor("SubprocessManager")

        if self._RUNNERS_FACTORY is None:
            self._RUNNERS_FACTORY = subprocess_manager.create_new_factory(
                "Libre Flow Tools"
            )
            self._register_runners()

        subprocess_manager.ensure_factory(self._RUNNERS_FACTORY)

    def get_factory(self):
        self.ensure_runners_loaded()
        return self._RUNNERS_FACTORY

    def get_entity_store(self):
        return self.admin.store
    
    def get_file_manager(self):
        return self.admin.file_manager
    
    def get_sync_manager(self):
        return self.synchronization.synchronize_files
    
    def touch(self):
        super(Project, self).touch()
        self.ensure_runners_loaded()
        self.update_user_environment()

    def _fill_ui(self, ui):
        if self._RUNNERS_FACTORY is None:
            self.ensure_runners_loaded()

        self.touch()

        if self.show_login_page():
            ui["custom_page"] = "libreflow.baseflow.LoginPageWidget"



import platform


def wrap_python_expr(expression):
    '''
    This functions wraps python expression for b3d within an exec()
    as windows shells are not handling well the line breaks
    '''
    _os = platform.system()
    if _os == "Windows":
        expression = 'exec(%s)' % repr(expression)
    return expression


from kabaret.app import plugin
from kabaret.app.ui.gui.widgets.flow.flow_view import FlowView
from kabaret.app.ui.gui.widgets.widget_view import QtWidgets, QtGui, QtCore
from kabaret.app import resources


class SearchResult(QtWidgets.QTreeWidgetItem):

    def __init__(self, tree, result):
        super(SearchResult, self).__init__(tree)
        self.session = tree.session
        self.set_result(result)
    
    def set_result(self, result):
        self._result = result
        self._update()
    
    def _update(self):
        self.setText(0, self.label())
        self.setIcon(0, self.icon())
    
    def id(self):
        return self._result['id']
    
    def goto_oid(self):
        return self._result['goto_oid']
    
    def label(self):
        return self.session.cmds.Flow.get_source_display(
            self.goto_oid()
        )
    
    def icon(self):
        thumbnail_info = self.session.cmds.Flow.get_thumbnail_info(
            self.goto_oid()
        )
        if thumbnail_info.get('is_resource', False):
            return resources.get_icon(
                (thumbnail_info['folder'], thumbnail_info['name'])
            )
        else:
            return resources.get_icon(('icons.flow', 'object'))


class NoMatchingResultsItem(QtWidgets.QTreeWidgetItem):

    def __init__(self, tree):
        super(NoMatchingResultsItem, self).__init__(tree)
        self.setText(0, 'No matching results')


class ResultList(QtWidgets.QTreeWidget):

    def __init__(self, search_bar):
        super(ResultList, self).__init__()

        self.search_bar = search_bar
        self.session = search_bar.session
        self.header().hide()

        self.itemClicked.connect(self.on_result_clicked)

        self._active_result = None
    
    def refresh(self):
        self.clear()
        self.selectionModel().clearSelection()

        results = self.session.cmds.Search.query_project_index(
            self.search_bar.current_project(),
            self.search_bar.query_filter()
        )
        
        if results:
            for res in results:
                SearchResult(self, res)
            
            self._active_result = 0
            self.update_active_result(0)
        else:
            self._active_result = None
            NoMatchingResultsItem(self)
    
    def active_result(self):
        selected = self.selectedItems()
        
        if not selected or self._active_result is None:
            return None
        else:
            return selected[0]
    
    def update_active_result(self, index):
        self.clearSelection()
        self.setCurrentItem(self.topLevelItem(index), 0, QtCore.QItemSelectionModel.Select)
    
    def incr_active_result(self):
        if self._active_result is not None and self._active_result < self.topLevelItemCount() - 1:
            self._active_result += 1
            self.update_active_result(self._active_result)
    
    def decr_active_result(self):
        if self._active_result is not None and self._active_result > 0:
            self._active_result -= 1
            self.update_active_result(self._active_result)
    
    def on_result_clicked(self, item, column):
        self.search_bar.view.goto_request(item.goto_oid())
        self.search_bar.search_input.lineedit_input.clearFocus()
    
    def resizeEvent(self, event):
        # From https://stackoverflow.com/a/37724553
        height = 2 * self.frameWidth() # border around tree
        it = QtWidgets.QTreeWidgetItemIterator(self)

        while it.value() is not None:
            index = self.indexFromItem(it.value())
            height += self.rowHeight(index)
            it += 1

        self.parentWidget().parentWidget().resize(self.parentWidget().parentWidget().width(), height)
        super(ResultList, self).resizeEvent(event)


class ResultListWidget(QtWidgets.QWidget):
    
    def __init__(self, search_bar):
        super(ResultListWidget, self).__init__()

        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.setMaximumHeight(200)

        self.result_list = ResultList(search_bar)
        lo = QtWidgets.QHBoxLayout()
        lo.addWidget(self.result_list)
        lo.setSpacing(0)
        lo.setContentsMargins(0, 0, 0, 0)
        self.setLayout(lo)

    def sizeHint(self):
        return QtCore.QSize(300, 200)


class ResultListDockWidget(QtWidgets.QDockWidget):

    def __init__(self, parent, search_bar):
        super(ResultListDockWidget, self).__init__(parent)

        self.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)
        self.setFloating(True)
        # Make title bar invisible (https://doc.qt.io/qt-5/qdockwidget.html#appearance)
        self.setTitleBarWidget(QtWidgets.QWidget())

        self.result_list_widget = ResultListWidget(search_bar)
        self.setWidget(self.result_list_widget)

        self.hide()


class SearchInputLineEdit(QtWidgets.QLineEdit):

    def __init__(self, search_input):
        super(SearchInputLineEdit, self).__init__()
        self.search_input = search_input
        self.setPlaceholderText('Search...')

        # Install callbacks
        self.textChanged.connect(self.search_input.refresh_result_list)
        self.returnPressed.connect(self.search_input.goto_active_result)

    def resizeEvent(self, event):
        self.search_input.update_result_list_width(self.width())
    
    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Escape:
            self.search_input.result_list.setVisible(False)
            self.clearFocus()
        elif event.key() == QtCore.Qt.Key_Up:
            self.search_input.decr_active_result()
        elif event.key() == QtCore.Qt.Key_Down:
            self.search_input.incr_active_result()
        else:
            super(SearchInputLineEdit, self).keyPressEvent(event)
    
    def focusInEvent(self, event):
        self.search_input.result_list.setVisible(True)
        g = self.search_input.result_list.geometry()
        g.setTopLeft(self.mapToGlobal(self.geometry().bottomLeft()))
        self.search_input.result_list.setGeometry(g)
        self.search_input.update_result_list_width(self.width())
        super(SearchInputLineEdit, self).focusInEvent(event)
    
    def focusOutEvent(self, event):
        self.search_input.result_list.setVisible(False)
        super(SearchInputLineEdit, self).focusOutEvent(event)


class SearchInputWidget(QtWidgets.QWidget):

    def __init__(self, search_bar):
        super(SearchInputWidget, self).__init__()
        self.search_bar = search_bar

        self.lineedit_input = SearchInputLineEdit(self)
        self.result_list = ResultListDockWidget(self, search_bar)

        vlo = QtWidgets.QVBoxLayout()
        vlo.addWidget(self.lineedit_input)
        vlo.setSpacing(0)
        vlo.setContentsMargins(0, 0, 0, 0)
        self.setLayout(vlo)
    
    def update_result_list_width(self, w):
        self.result_list.resize(w, self.result_list.height())
    
    def refresh_result_list(self):
        self.result_list.result_list_widget.result_list.refresh()
    
    def goto_active_result(self):
        active = self.result_list.result_list_widget.result_list.active_result()
        
        if active is not None:
            self.search_bar.view.goto_request(active.goto_oid())
            self.lineedit_input.clearFocus()
    
    def incr_active_result(self):
        self.result_list.result_list_widget.result_list.incr_active_result()
    
    def decr_active_result(self):
        self.result_list.result_list_widget.result_list.decr_active_result()


class SearchBar(QtWidgets.QWidget):

    def __init__(self, parent, view, session):
        super(SearchBar, self).__init__(parent)
        self.view = view
        self.session = session
        
        self.combobox_projects = QtWidgets.QComboBox()
        self.search_input = SearchInputWidget(self)

        hlo = QtWidgets.QHBoxLayout()
        hlo.addWidget(self.combobox_projects)
        hlo.addWidget(self.search_input)
        hlo.setSpacing(2)
        hlo.setContentsMargins(2, 0, 2, 0)
        self.setLayout(hlo)

        self.refresh()
    
    def refresh(self):
        self.combobox_projects.insertItems(0, self.project_names())
    
    def current_project(self):
        return self.combobox_projects.currentText() or None
    
    def query_filter(self):
        return self.search_input.lineedit_input.text() or None
    
    def project_names(self):
        return self.session.cmds.Search.list_project_names()


class SearchFlowView(FlowView):

    def __init__(self, session, view_id=None, hidden=False, area=None, oid=None, root_oid=None):
        self.search_bar = None
        super(SearchFlowView, self).__init__(session, view_id=view_id, hidden=hidden, area=area, oid=oid, root_oid=root_oid)

    def build_top(self, top_parent, top_layout, header_parent, header_layout):
        super(SearchFlowView, self).build_top(top_parent, top_layout, header_parent, header_layout)
        self.search_bar = SearchBar(top_parent, self, self.session)
        top_layout.addWidget(self.search_bar)


class SearchFlowViewPlugin:
    """
    Custom Flow view.

    Will only be installed if no other view
    is registered under the "Flow" view type name.
    """

    @plugin(trylast=True)
    def install_views(session):
        if not session.is_gui():
            return

        type_name = SearchFlowView.view_type_name()
        if not session.has_view_type(type_name):
            session.register_view_type(SearchFlowView)
            session.add_view(type_name)



import os
import subprocess
import pathlib
import time
import uuid
from datetime import datetime

import logging

import kabaret.subprocess_manager as ksm

logger = logging.getLogger(__name__)


class Runner(ksm.runner_factory.Runner):
    
    def __init__(self, version=None, label=None, extra_argv=[], extra_env={}):
        super(Runner, self).__init__(
            version=version,
            label=label,
            extra_argv=extra_argv,
            extra_env=extra_env,
        )
        self._last_run_time = None
        self._last_cmd = None

    def run(self):
        cmd = [self.executable()]
        cmd.extend(self.argv())

        env = self.env()

        os_flags = {}

        # Disowning processes in linux/mac
        if hasattr(os, "setsid"):
            os_flags["preexec_fn"] = os.setsid

        # Disowning processes in windows
        if hasattr(subprocess, "STARTUPINFO"):
            # Detach the process
            os_flags["creationflags"] = subprocess.CREATE_NEW_CONSOLE

            # Hide the process console
            startupinfo = subprocess.STARTUPINFO()
            if self.show_terminal():
                flag = "/C"
                if self.keep_terminal():
                    flag = "/K"
                cmd = ["cmd", flag] + cmd
            else:
                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW

            os_flags["startupinfo"] = startupinfo

        logger.info("Running Subprocess: %r", cmd)
        
        if not os.path.exists(self._get_log_dir()):
            os.mkdir(self._get_log_dir())
        
        # Store run time used to compute log paths
        self._last_run_time = time.time()
        cmd = [str(arg) for arg in cmd]
        self._last_cmd = ' '.join(cmd)

        with open(self.get_log_path(), 'w+') as log_fd:
            self._popen = subprocess.Popen(
                cmd,
                env=env,
                stdout=log_fd,
                stderr=log_fd,
                **os_flags,
            )
    
    def _get_log_dir(self):
        return str(pathlib.Path.home()) + "/.libreflow/log/"

    def get_log_path(self):
        dt = datetime.fromtimestamp(self._last_run_time)
        dt = dt.astimezone().strftime("%Y-%m-%dT%H-%M-%S%z")
        
        path = os.path.join(
            self._get_log_dir(),
            '%s_%s.log' % (self.runner_name(), dt),
        )
        return path
    
    def has_run(self):
        return self._last_run_time is not None
    
    def last_run_time(self):
        return self._last_run_time
    
    def cmd(self):
        return self._last_cmd
    
    def pid(self):
        return self._popen.pid
    
    def last_run_time(self):
        return self._last_run_time
    
    def terminate(self):
        self._popen.terminate()
    
    def kill(self):
        self._popen.kill()

    def is_running(self):
        return self._popen.poll() is None
    
    def return_code(self):
        return self._popen.returncode


class SubprocessManager(object):
    """
    The SubprocessManager manages a list of Runner instances.
    
    From original kabaret.subprocess_manager.runner_factory.SubprocessManager
    """

    def __init__(self):
        super(SubprocessManager, self).__init__()
        self._runners = {}

    @staticmethod
    def _get_runner_info(runner, rid):
        return dict(
            id=rid,
            label=runner.label,
            name=runner.runner_name(),
            icon=runner.runner_icon(),
            version=runner.version,
            is_running=runner.is_running(),
            log_path=runner.get_log_path(),
            command=runner.cmd(),
            last_run_time=runner.last_run_time(),
            pid=runner.pid(),
        )
    
    @staticmethod
    def _get_unknown_runner_info(rid):
        return dict(
            id=rid,
            label='!!!',
            name='!!!',
            icon=('icons.libreflow', 'exclamation-sign-colored'),
            version='!!!',
            is_running=False,
            log_path='!!!',
            command='!!!',
            last_run_time=-1,
            pid=-1,
        )
    
    def get_runner_infos(self):
        """
        Return a list of dict with keys:
            label, name, icon, version, is_running, log_path
        """
        infos = []
        for id, runner in self._runners.items():
            if runner.has_run():
                infos.append(
                    self._get_runner_info(runner, id)
                )
        
        return infos
    
    def get_runner_info(self, rid):
        """
        Returns data of the runner indexed with
        the given id `rid`, as a dict with keys:
            id, label, name, icon, version, is_running, log_path, pid
        """
        try:
            runner = self._runners[rid]
        except KeyError:
            return self._get_unknown_runner_info(rid)
        else:
            return self._get_runner_info(runner, rid)
    
    def get_runner(self, rid):
        return self._runners.get(rid, None)
    
    def delete_runner(self, rid):
        try:
            self._runners.pop(rid)
        except KeyError:
            return False
        else:
            return True

    def run(self, runner):
        rid = self._add_runner(runner)
        runner.run()
        return rid

    def _add_runner(self, runner):
        rid = str(uuid.uuid4())
        max_tries = 10
        i = 0
        while rid in self._runners and i < max_tries:
            rid = str(uuid.uuid4())
            i += 1
        if i == max_tries:
            raise Exception('Could not create a new runner...')

        self._runners[rid] = runner
        return rid



import re
import os
from pathlib import Path
import shutil
from distutils.dir_util import copy_tree

from kabaret import flow
from kabaret.flow_contextual_dict import get_contextual_dict
from kabaret.subprocess_manager.flow import RunAction

from libreflow import baseflow


def remove_files(root_path):
    """
    Removes the files contained in a folder with a given path.
    https://stackoverflow.com/a/185941
    """
    if not os.path.exists(root_path):
        print('Failed to delete %s. Reason: Folder does not exists' % (root_path))
        return
    for filename in os.listdir(root_path):
        file_path = os.path.join(root_path, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            # elif os.path.isdir(file_path):
            #     shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))


def move_files(src_path, dst_path):
    for filename in os.listdir(src_path):
        src_file_path = os.path.join(src_path, filename)
        dst_file_path = os.path.join(dst_path, filename)
        try:
            if os.path.isfile(src_file_path) or os.path.islink(src_file_path):
                shutil.move(src_file_path, dst_file_path)
        except Exception as e:
            print('Failed to move %s to %s. Reason: %s' % (src_file_path, dst_file_path, e))


class RenderSettingsTemplate(flow.values.ChoiceValue):

    CHOICES = ["majorque_rendu_preview", "majorque_rendu_output"]


class OutputModuleTemplate(flow.values.ChoiceValue):

    CHOICES = ["majorque_preview", "majorque_output"]


class RenderWithAfterEffect(RunAction):

    _revision = flow.Parent()
    _files = flow.Parent(5)

    template_label = flow.Label("Templates:")
    render_settings = flow.Param("majorque_rendu_output", RenderSettingsTemplate)
    output_module = flow.Param("majorque_output", OutputModuleTemplate)

    def get_buttons(self):
        return ["Render", "Cancel"]

    def runner_name_and_tags(self):
        return "AfterEffectRender", []

    def get_run_label(self):
        return 'Render'

    @classmethod
    def supported_extensions(cls):
        return ["aep"]
    
    def allow_context(self, context):
        return (
            context
            and self._revision._file.format.get() in self.supported_extensions()
        )


class MakeAfterEffectPreview(RenderWithAfterEffect):

    render_settings = flow.Param("majorque_rendu_preview", RenderSettingsTemplate)
    output_module = flow.Param("majorque_preview", OutputModuleTemplate)

    def extra_argv(self):
        settings = get_contextual_dict(self._revision, "settings")
        file = self._revision._file

        return [
            "-project", self._revision.get_path(),
            "-comp", "%s_%s_output" % (settings["sequence"], settings["shot"]),
            # "-s", "%s" % start,
            # "-e", "%s" % end,
            "-RStemplate", self.render_settings.get(),
            # "-OMtemplate", self.output_module.get(),
            "-output", "%s/%s_%s_anim_%s.mov" % (
                file.get_playblast_folder(),
                settings["sequence"],
                settings["shot"],
                self._revision.name(),
            ),
        ]
    
    def run(self, button):
        if button == "Cancel":
            return

        if not self._files.has_mapped_name("preview"):
            self._files.add_folder("preview")
            self._files.touch()
        
        super(MakeAfterEffectPreview, self).run(button)


class MakeAfterEffectRender(RenderWithAfterEffect):

    def extra_argv(self):
        settings = get_contextual_dict(self._revision, "settings")
        file = self._revision._file

        return [
            "-project", self._revision.get_path(),
            "-comp", "%s_%s_output" % (settings["sequence"], settings["shot"]),
            # "-s", "%s" % start,
            # "-e", "%s" % end,
            "-RStemplate", self.render_settings.get(),
            "-OMtemplate", self.output_module.get(),
            "-output", "%s/%s_%s_anim.[####].png" % (
                file.get_render_path(),
                settings["sequence"],
                settings["shot"],
            ),
        ]
    
    def run(self, button):
        if button == "Cancel":
            return

        render_path = self._revision._file.get_render_path()

        if not self._files.has_mapped_name("render"):
            self._files.add_folder("render")
            self._files.touch()
        if not os.path.exists("%s/tmp" % render_path):
            os.makedirs("%s/tmp" % render_path)
        if not os.path.exists("%s/previous" % render_path):
            os.makedirs("%s/previous" % render_path)
        
        remove_files("%s/previous" % render_path)
        move_files(render_path, "%s/previous" % render_path)
        
        super(MakeAfterEffectRender, self).run(button)


class Revision(baseflow.file.Revision):

    playblast_path = flow.Computed()

    make_preview = flow.Child(MakeAfterEffectPreview)
    make_render = flow.Child(MakeAfterEffectRender)

    def has_playblast(self):
        return os.path.exists(self.playblast_path.get())

    def compute_child_value(self, child_value):
        if child_value is self.playblast_path:
            child_value.set(
                os.path.join(
                    self._file.get_playblast_folder(),
                    "%s_%s-movie.mov" % (self._file.complete_name.get(), self.name()),
                )
            )
        else:
            super(Revision, self).compute_child_value(child_value)


class Revisions(baseflow.file.Revisions):
    @classmethod
    def mapped_type(cls):
        return Revision


class CreateTrackedFileAction(baseflow.file.CreateTrackedFileAction):
    def run(self, button):
        if button == "Cancel":
            return

        settings = get_contextual_dict(self, "settings")
        file_category = settings.get("file_category", None)

        name = self.file_name.get()
        prefix = ""

        if file_category is not None:
            if file_category == "PROD":
                prefix = "{episode}_{sequence}_{shot}_{department}_"
            elif file_category == "LIB":
                prefix = "{asset_name}_{department}_"

            prefix = prefix.format(**settings)

        self.root().session().log_debug(
            "Creating file %s.%s" % (name, self.file_format.get())
        )

        self._files.create_file(name, extension=self.file_format.get(), base_name=prefix+name)
        self._files.touch()


class PublishAndRenderPlayblast(baseflow.file.PublishFileAction):
    def run(self, button):
        if button == "Cancel":
            return self.get_result(next_action=self._file.render_blender_playblast.oid())

        super(PublishAndRenderPlayblast, self).run(button)
        published_revision = self._file.get_head_revision()

        self._file.render_blender_playblast.revision_name.set(published_revision.name())

        return self._file.render_blender_playblast.run("Render")


class TrackedFile(baseflow.file.TrackedFile):

    with flow.group("Playblast"):
        render_blender_playblast = flow.Child(baseflow.file.RenderBlenderPlayblast)
        publish_and_playblast_blender = flow.Child(PublishAndRenderPlayblast)

    with flow.group("Open with"):
        open_with_blender = flow.Child(baseflow.file.OpenWithBlenderAction).ui(
            label="Blender"
        )
        open_with_krita = flow.Child(baseflow.file.OpenWithKritaAction).ui(
            label="Krita"
        )
        open_with_vscodium = flow.Child(baseflow.file.OpenWithVSCodiumAction).ui(
            label="VSCodium"
        )
        open_with_notepadpp = flow.Child(baseflow.file.OpenWithNotepadPPAction).ui(
            label="Notepad++"
        )

    def has_playblast(self):
        for rev in self.get_revisions().mapped_items():
            if rev.has_playblast():
                return True

        return False

    def get_playblast_folder(self):
        return os.path.join(
            self.root().project().get_root(),
            self._department.path.get(),
            "preview"
        )
    
    def get_render_path(self):
        return os.path.join(
            self.root().project().get_root(),
            self._department.path.get(),
            "render"
        )


class FileSystemMap(baseflow.file.FileSystemMap):
    def add_tracked_file(self, name, extension, complete_name):
        key = "%s_%s" % (name, extension)
        file = self.add(key, object_type=TrackedFile)
        file.format.set(extension)
        file.complete_name.set(complete_name)

        # Create file folder
        try:
            self.root().session().log_debug(
                "Create file folder '{}'".format(file.get_path())
            )
            os.makedirs(file.get_path())
        except OSError:
            self.root().session().log_error(
                "Creation of file folder '{}' failed.".format(file.get_path())
            )
            pass

        # Create current revision folder
        current_revision_folder = os.path.join(file.get_path(), "current")

        try:
            self.root().session().log_debug(
                "Create current revision folder '{}'".format(
                    current_revision_folder
                )
            )
            os.mkdir(current_revision_folder)
        except OSError:
            self.root().session().log_error(
                "Creation of current revision folder '{}' failed".format(
                    current_revision_folder
                )
            )
            pass

        return file



'''

    This module is a main script for the flow jobs.
    It instanciates a JobWorkerSession and exexutes the job pointed by the env-var "KABARET_JOBS_FLOW_OID"


'''

from __future__ import print_function

import sys
import os
import traceback

from kabaret.app.session import KabaretSession
import libreflow.utils.kabaret as kutils
from libreflow.utils.kabaret.jobs import jobs_actor
from libreflow.baseflow.runners import (
    Blender,
    AfterEffectsRender,
    SessionWorker,
    MarkSequenceRunner,
)

from libreflow.resources import file_templates


def try_initialize_maya():
    try:
        import maya.standalone
    except ImportError:
        pass
    else:
        print('This looks like a maya pythong, let us initialize it.')
        try:
            maya.standalone.initialize()
        except:
            print('!!! Error initializing Maya:')
            traceback.print_exc()
            print('!!! Error initializing Maya.')
            sys.exit(-1)
        print('Maya initialization done.')


class JobsWorkerSession(KabaretSession):

    def __init__(self):
        super(JobsWorkerSession, self).__init__('JobsWorker')

        self.cmds.Cluster.connect_from_env()
        self.execute_job()

    def _create_actors(self):
        super(JobsWorkerSession, self)._create_actors()
        self.jobs_actor = jobs_actor.Jobs(self)
        
        subprocess_manager = kutils.subprocess_manager.SubprocessManager(self)

        factory = subprocess_manager.create_new_factory('JobsWorker Tools')
        factory.ensure_runner_type(Blender)
        factory.ensure_runner_type(AfterEffectsRender)
        factory.ensure_runner_type(MarkSequenceRunner)
        factory.ensure_runner_type(SessionWorker)

        subprocess_manager.ensure_factory(factory)

    def execute_job(self):
        self.jobs_actor.worker_execute_job()


if __name__ == '__main__':
    try_initialize_maya()
    JobsWorkerSession()


from kabaret.subprocess_manager.flow import RunAction as KabaretSMRunAction


class RunAction(KabaretSMRunAction):

    @classmethod
    def get_result(
            cls, close=None, refresh=None,
            goto=None, goto_target=None, goto_target_type=None, 
            next_action=None, runner_id=None
    ):
        ret = super(RunAction, cls).get_result(
            close=close, refresh=refresh,
            goto=goto, goto_target=goto_target, goto_target_type=goto_target_type,
            next_action=next_action
        )
        ret.update(dict(runner_id=runner_id))

        return ret

    def run(self, button):
        name, tags = self.runner_name_and_tags()
        rid = self.root().session().cmds.SubprocessManager.run(
            runner_name=name,
            tags=tags,
            version=self.get_version(button),
            label=self.get_run_label(),
            extra_argv=self.extra_argv(),
            extra_env=self.extra_env(),
        )
        return self.get_result(runner_id=rid)



from __future__ import print_function

import six
import os
import tempfile

from kabaret import flow

from . import jobs_flow

def make_default_temp_dir():
    '''
    Just a tool to have a kool default log path for the demo.
    Windows user have complicated tempdir, so Imake it shorter.
    '''
    fullpath = tempfile.gettempdir()
    drive, path = os.path.splitdrive(fullpath)
    if drive:
        fullpath = drive+os.path.sep

    return os.path.join(fullpath, 'KABARET_JOBS_DEMO')



class TaskJob(jobs_flow.Job):

    _task = flow.Parent(2)

    with flow.group('Task Parameters'):
        param_01 = flow.Param().ui(editable=False)
        param_02 = flow.Param().ui(editable=False)

    def get_log_filename(self):
        return os.path.join(self.root().project().log_path.get(), self._task.name(), self.name()+'.job_log')

    def _do_job(self):
        print('############# JOBS DEMO - TaskJob Starts #############')
        self._task.status.set('WIP')
        import time
        count = 5
        for i in range(count):
            self.set_progress(i, count, 'Working hard {}'.format(i))
            time.sleep(1)
        self._task.status.set('DONE')
        self.set_progress(count, count, 'Everything is Done !')
        print('############# JOBS DEMO - TaskJob Ends #############')
        

class SubmitTaskJobAction(jobs_flow.SubmitJobAction):

    with flow.group('Task Parameters'):
        job_param_01 = flow.Param()
        job_param_02 = flow.Param()

    sep = flow.Separator()

    def configure_job(self, job):
        job.param_01.set(self.job_param_01.get())
        job.param_02.set(self.job_param_02.get())

class TaskJobs(jobs_flow.Jobs):

    submit_job = flow.Child(SubmitTaskJobAction)

    @classmethod
    def job_type(cls):
        return TaskJob


class TaskStatusValue(flow.values.ChoiceValue):

    CHOICES = ('INV', 'WIP', 'DONE')

    @classmethod
    def default(cls):
        return cls.CHOICES[0]

class Task(flow.Object):

    status = flow.Param(TaskStatusValue.default(), TaskStatusValue)
    jobs = flow.Child(TaskJobs)

class CreateTaskAction(flow.Action):
    
    _tasks = flow.Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        self._tasks.create_new_task()

class Tasks(flow.Map):

    create_task = flow.Child(CreateTaskAction)

    @classmethod
    def mapped_type(cls):
        return Task

    def create_new_task(self):
        i = len(self)
        n = 'Task'+str(i).zfill(5)
        self.add(n)
        self.touch()

class JobsDemoFlow(flow.Object):

    log_path = flow.Param(make_default_temp_dir())
    exe = flow.Param('C:/Program Files/Sublime Text 3/sublime_text.exe')
    flags = flow.Param('')

    tasks = flow.Child(Tasks)




import time

from kabaret import flow
from kabaret.flow_entities.entities import Entity, EntityCollection


class CustomEntityCollection(EntityCollection):
    '''
    This class provides add and remove operators, and allows type injections
    on elements whose type is derived from its mapped_type (returned by
    `mapped_type()`).
    '''
    _mapped_types = flow.HashParam()

    def get_property(self, entity_name, property_name):
        value = (
            self.get_entity_store()
            .get_collection(self.collection_name())
            .find_one(
                {'name': entity_name},
                { property_name: 1},
            )
        )
        try:
            return value[property_name]
        except KeyError:
            default = getattr(self._get_mapped_item_type(entity_name), property_name).get_default_value()
            return default

    def add(self, name, object_type=None):
        '''
        Adds an object to the map.
        If provided, object_type must be a subclass of the map's mapped_type (returned by the
        classmethod mapped_type())
        '''
        mapped_type = flow.injection.resolve(
            self.mapped_type(), self
        )
        if object_type is None:
            object_type = mapped_type
        elif not issubclass(object_type, mapped_type):
            raise TypeError(
                'Cannot add %r of type %r to Map %r: not a subclass of %r' % (
                    name, object_type, self.oid(), mapped_type
                )
            )
        else:
            # Resolve potential injections on provided object type
            object_type = flow.injection.resolve(object_type, self)

        if '.' in name:
            raise TypeError(
                'Invalid object name %r (it must be a valid attribute name).' % (name,))
        try:
            exec(name + '=None') in {}
        except:
            raise TypeError(
                'Invalid object name %r (it must be a valid attribute name).' % (name,))
        if name in dir(self):
            raise ValueError(
                'Cannot add an item "%r", this name is already defined in the class "%s" (%s).' %
                (
                    name, self.__class__.__name__, self._mng.oid()
                )
            )
        if self.has_mapped_name(name):
            raise ValueError(
                'An item %r is already mapped in %r.' %
                (
                    name, self._mng.oid()
                )
            )

        object_qualified_type_name = self._mng.get_qualified_type_name(
            object_type)
        
        c = self.get_entity_store().get_collection(self.collection_name())
        c.insert_one({'name': name})
        self._mapped_types.set_key(name, object_qualified_type_name)

        # Reset collection cache to get newly created entry
        self._document_cache = None

        return self.get_mapped(name)

    def remove(self, name):
        self._mapped_types.del_key(name)
        c = self.get_entity_store().get_collection(self.collection_name())
        c.delete_one({'name': name})

    def _get_mapped_item_type(self, mapped_name):
        '''
        Returns the type of the given mapped item.
        NB: The GUI assumes that the returned type is a subclass
        of mapped_type().

        '''
        object_qualified_type_name = self._mapped_types.get_key(mapped_name)
        try:
            object_type = self._mng.qualified_type_name_to_type(
                object_qualified_type_name
            )
        except ImportError:
            mapped_type = flow.injection.resolve(
                self.mapped_type(), self
            )
            logging.getLogger('kabaret.ui').debug(
                'WARNING: unable to access type %r. Downgrading to map default type %r' % (
                    object_qualified_type_name, mapped_type
                )
            )
            object_type = mapped_type
        return object_type


class GlobalEntityCollection(EntityCollection):
    '''
    An entity collection which stores entities at the scope of an entire project.

    Entities of this collection are mapped by their oids.
    '''
    
    def columns(self):
        return ['Oid', 'Type']
    
    def _fill_row_cells(self, row, item):
        self.mapped_names()

        oid = item.oid()
        row['Oid'] = oid
        row['Type'] = self._document_cache[oid].get(
            'mapped_type',
            self._mng.get_qualified_type_name(self.mapped_type())
        )
    
    def get_mapped(self, name):
        return self.root().get_object(name)


class EntityView(EntityCollection):
    '''
    A collection acting as a view on a subset of a GlobalEntityCollection.

    User may return the global collection name in the `collection_name()` method.
    '''

    def mapped_names(self, page_num=0, page_size=None):
        oids = super(EntityView, self).mapped_names(page_num, page_size)
        return [oid.rsplit('/', maxsplit=1)[-1] for oid in oids]

    def query_filter(self):
        return {'name': {'$regex': f'^{self.oid()}/[^/]*'}}
    
    def add(self, name, object_type=None):
        '''
        Adds an object to the map.
        If provided, object_type must be a subclass of the map's mapped_type (returned by the
        classmethod mapped_type())
        '''
        mapped_type = flow.injection.resolve(
            self.mapped_type(), self
        )

        # Check file type
        if object_type is None:
            object_type = mapped_type
        elif not issubclass(object_type, mapped_type):
            raise TypeError(
                'Cannot add %r of type %r to Map %r: not a subclass of %r' % (
                    name, object_type, self.oid(), mapped_type
                )
            )

        # Check file name
        if '.' in name:
            raise TypeError(
                'Invalid object name %r (it must be a valid attribute name).' % (name,))
        try:
            exec(name + '=None') in {}
        except:
            raise TypeError(
                'Invalid object name %r (it must be a valid attribute name).' % (name,))
        if name in dir(self):
            raise ValueError(
                'Cannot add an item "%r", this name is already defined in the class "%s" (%s).' %
                (
                    name, self.__class__.__name__, self._mng.oid()
                )
            )
        
        if self.has_mapped_name(name):
            raise ValueError(
                'An item %r is already mapped in %r.' %
                (
                    name, self._mng.oid()
                )
            )

        object_qualified_type_name = self._mng.get_qualified_type_name(object_type)
        oid = f'{self.oid()}/{name}'

        # Add entry to project's Mongo store
        self.get_entity_store().get_collection(self.collection_name()).insert_one({
            'name': oid,
            'mapped_type': object_qualified_type_name
        })
        self._document_cache = None # Reset cache

        return self.get_mapped(name)

    def remove(self, name):
        '''
        Removes an Entity from the map.
        '''
        try:
            self._mng.destroy_child(name)
        except flow.MissingChildError:
            # was not yet instantiated, nothing to destroy
            pass

        self.get_entity_store().get_collection(self.collection_name()).delete_one(
            {'name': f'{self.oid()}/{name}'}
        )
        self._document_cache = None # Reset cache
    
    def clear(self):
        '''
        Clears all Entities belonging to this map.
        '''
        for name in self.mapped_names():
            try:
                self._mng.destroy_child(name)
            except flow.MissingChildError:
                # was not yet instantiated, nothing to destroy
                pass
        
        self.get_entity_store().get_collection(self.collection_name()).delete_many(
            self.query_filter()
        )
        self._document_cache = None # Reset cache
    
    def set_property(self, entity_name, property_name, value):
        self.get_entity_store().get_collection(self.collection_name()).update_one(
            {"name": f'{self.oid()}/{entity_name}'},
            {"$set": {property_name: value}},
        )

    def get_property(self, entity_name, property_name):
        self.root().session().log_debug(f'===========> {self.oid()}/{entity_name} {property_name}')
        value = (
            self.get_entity_store()
            .get_collection(self.collection_name())
            .find_one(
                {"name": f'{self.oid()}/{entity_name}'},
                {property_name: 1},
            )
        )
        try:
            return value[property_name]
        except KeyError:
            default = getattr(self._get_mapped_item_type(entity_name), property_name).get_default_value()
            return default
    
    def _fill_row_cells(self, row, item):
        if self._document_cache is None:
            # trigger a cache update:
            # FIXME: this cache update is not paginated :'(
            self.mapped_names()
        entity_oid = item.oid()
        for name, label in self.mapped_type().get_property_names_and_labels():
            row[label] = self._document_cache[entity_oid].get(name, "")
        row['Name'] = item.name()

    def _get_mapped_item_type(self, mapped_name):
        self.mapped_names()
        
        oid = f'{self.oid()}/{mapped_name}'
        object_qualified_type_name = self._document_cache[oid].get('mapped_type', None)
        
        if object_qualified_type_name is None:
            return self.mapped_type()
        
        try:
            object_type = self._mng.qualified_type_name_to_type(
                object_qualified_type_name
            )
        except ImportError:
            object_type = self.mapped_type()
            self.root().session().log_debug(
                'WARNING: unable to access type %r. Downgrading to map default type %r' % (
                    object_qualified_type_name, object_type
                )
            )
        
        return object_type



import sys
import os
import argparse

os.environ["QT_MAC_WANTS_LAYER"] = '1'

from kabaret.app.ui import gui
from kabaret.app.ui.gui.styles import Style
from kabaret.app.ui.gui.styles.gray import GrayStyle
from qtpy import QtWidgets, QtGui, QtCore
from kabaret.subprocess_manager import SubprocessManager

from .session import BaseGUISession, DebugGUISession
from .resources.icons import libreflow
from .resources import file_templates

CUSTOM_HOME = True
if CUSTOM_HOME:
    from kabaret.app.actors.flow import Flow
    from .custom_home import MyHomeRoot

DEBUG = False
SCRIPT_VIEW = True
try:
    from kabaret.script_view import script_view
except ImportError:
    SCRIPT_VIEW = False

GrayStyle()


class MyStudioGUISession(gui.KabaretStandaloneGUISession):
    def register_view_types(self):
        super(MyStudioGUISession, self).register_view_types()

        # type_name = self.register_view_type(SessionToolBar)
        # self.add_view(type_name)
        if SCRIPT_VIEW:
            type_name = self.register_view_type(script_view.ScriptView)
            self.add_view(
                type_name, hidden=not DEBUG, area=QtCore.Qt.RightDockWidgetArea
            )

    def _create_actors(self):
        """
        Instanciate the session actors.
        Subclasses can override this to install customs actors or
        replace default ones.
        """
        if CUSTOM_HOME:
            Flow(self, CustomHomeRootType=MyHomeRoot)
        else:
            return super(MyStudioGUISession, self)._create_actors()
        subprocess_manager = SubprocessManager(self)


def process_remaining_args(args):
    parser = argparse.ArgumentParser(
        description='Libreflow Session Arguments'
    )
    parser.add_argument(
        '-u', '--user', dest='user'
    )
    parser.add_argument(
        '-s', '--site', default='LFS', dest='site'
    )
    values, _ = parser.parse_known_args(args)

    if values.site:
        os.environ["KABARET_SITE_NAME"] = values.site
    if values.user:
        os.environ["USER_NAME"] = values.user


def main(argv):
    (
        session_name,
        host,
        port,
        cluster_name,
        db,
        password,
        debug,
        remaining_args,
    ) = MyStudioGUISession.parse_command_line_args(argv)

    session = MyStudioGUISession(session_name=session_name, debug=debug)
    session.cmds.Cluster.connect(host, port, cluster_name, db, password)

    process_remaining_args(remaining_args)

    session.start()
    session.close()


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse
import os

from libreflow.session import JobsNodeSession


def process_remaining_args(args):
    parser = argparse.ArgumentParser(
        description='Libreflow Session Arguments'
    )
    parser.add_argument(
        '-r', '--root_dir', dest='root_dir'
    )
    parser.add_argument(
        '-s', '--site', default='LFS', dest='site'
    )
    parser.add_argument(
        '--blender_exec_path', dest='blender_exec_path'
    )
    values, remaining_args = parser.parse_known_args(args)

    if values.root_dir:
        os.environ["ROOT_DIR"] = values.root_dir
    if values.site:
        os.environ["KABARET_SITE_NAME"] = values.site
    if values.blender_exec_path:
        os.environ["BLENDER_EXEC_PATH"] = values.blender_exec_path
    
    return remaining_args


if __name__ == '__main__':
    argv = sys.argv[1:]  # get ride of first args wich is script filename
    session_name, host, port, cluster_name, db, password, debug, remaining_args \
        = JobsNodeSession.parse_command_line_args(argv)
    pools = process_remaining_args(remaining_args)
    if not pools:
        print('\n!!!\nUsage:', sys.argv[0], 'pool_name pool_name ...\n\n')
    else:
        node = JobsNodeSession(pools, session_name=session_name, debug=debug)
        node.cmds.Cluster.connect(host, port, cluster_name, db, password)

        node.wait_for_jobs()
        print('Jobs Node closed.')
        



from __future__ import print_function

import time
import traceback

from kabaret import flow


class Job(flow.Object):
    ''' 
    This is the base class for Flow jobs.

    You must subclass it and implement:

        get_log_filename():
            Return the filename of the log file.
            (make it quick and stable !)

        _do_job(self):
            Do the job you need to do, periodically calling:
                self.set_progress()
                self.show_message()
                self.on_error()
            Use simple print call to write to the log file.
            
    You may subclass it and override:
        get_worker_interpreter() => to use another python (mayapy, hython, ...)
        get_worker_filename() => to use you own worker (if you're using custom actors in your flow)

    '''
    _jobs = flow.Parent()

    job_id = flow.Param(None).ui(editable=False)
    owner = flow.Computed()
    status = flow.Computed()
    created_on = flow.Computed()
    icon = flow.Computed()
    log_filename = flow.Computed()
    progress = flow.Param(0).ui(editor='percent', editable=False)
    message = flow.Param('').ui(editor='textarea', editable=False)

    def compute_child_value(self, child_value):
        if child_value is self.owner:
            self.owner.set(self.get_owner())
        elif child_value in (self.status, self.icon):
            self._compute_status_and_icon()
        elif child_value is self.created_on:
            self.created_on.set(self.get_created_on())
        elif child_value is self.log_filename:
            self.log_filename.set(self.get_log_filename())

    def touch(self):
        self.status.touch()
        self.icon.touch()
        self._jobs.touch()

    def get_job(self):
        return self.root().session().cmds.Jobs.get_job_info(self.job_id.get())

    def get_owner(self):
        return self.get_job().get('owner', '???')

    def get_created_on(self):
        return self.get_job().get('created_on', '???')

    def _compute_status_and_icon(self):
        job = self.get_job()
        self.status.set(job.get('status', '?!?'))
        self.icon.set(job.get('icon_ref', None))

    def get_log_filename(self):
        raise NotImplementedError('You must return the filename of the log file.')

    def get_worker_interpreter(self):
        '''
        Subclasses may override this to specify a arbitrary python interpreter to
        use to exectute the job.

        Default is to return None, which leads to using current interpretor.
        '''
        return None

    def get_worker_filename(self):
        '''
        Subclasses may override this to specify a arbitrary filename to
        use to as worker for the job.

        Default is to return None, which leads to using kabaret.jobs.jobs_flow_worker.
        '''
        return None

    def submit(self, pool, priority, label, creator, owner, paused, show_console):
        jid = self.root().session().cmds.Jobs.create_flow_job(
            oid=self.oid(),
            pool=pool,
            priority=priority,
            paused=paused,
            label=label,
            creator=creator,
            owner=owner,
            show_console=show_console,
            interpreter=self.get_worker_interpreter(),
            worker_filename=self.get_worker_filename(),
        )
        self.job_id.set(jid)

    def execute(self):
        print('----------------EXECUTING JOB', self.oid())
        self.touch()
        self.root().session().cmds.Jobs.set_job_in_progress(self.job_id.get())
        try:
            self._do_job()
        except Exception as err:
            self.on_error(traceback.format_exc())
        else:
            self.root().session().cmds.Jobs.set_job_done(self.job_id.get())
            self.status.touch()
            self.touch()
        finally:
            self.touch()

    def set_progress(self, step_index, step_count, step_message):
        percent = (step_index*100.0/step_count)
        self.progress.set(percent)
        print('[PROGRESS] {:.2f}%'.format(percent))
        message = '{}/{}: {}'.format(step_index, step_count, step_message)
        self.show_message(message)
        self.touch()

    def show_message(self, message):
        print(message)
        self.message.set(message)
        self.touch()

    def on_error(self, error_message):
        self.show_message('!!! ERROR: {}'.format(error_message))
        self.root().session().cmds.Jobs.set_job_error(self.job_id.get(), error_message)
        self.status.touch()
        self.touch()

    def _do_job(self):
        '''
        Override this to implement your job.
        (See constructor doc)
        '''
        raise NotImplementedError('You must implement the job to do here.')


class SubmitJobAction(flow.Action):

    _jobs = flow.Parent()

    with flow.group('Job Dispatch'):
        pool = flow.Param('Farm')
        priority = flow.IntParam(100)
        paused = flow.BoolParam(True)
        show_console = flow.BoolParam(True)

    def needs_dialog(self):
        return True

    def get_buttons(self):
        return ['Submit', 'Submit and Open']

    def run(self, button):
        job = self._jobs.create_job()
        self.configure_job(job)
        self._jobs.touch()
        job.submit(
            self.pool.get(),
            self.priority.get(),
            self.paused.get(),
            self.show_console.get(),
        )

        if 'Open' in button:
            return self.get_result(goto=job.oid())
        return 

    def configure_job(self, job):
        pass

class Jobs(flow.Map):

    _next_id = flow.IntParam(0)

    @classmethod
    def job_type(cls):
        raise NotImplementedError('This should return your subclass of Job')

    @classmethod
    def mapped_type(cls):
        return cls.job_type()

    def mapped_names(self, *args, **kwargs):
        names = super(Jobs, self).mapped_names(*args, **kwargs)
        return reversed(names)

    def _get_next_job_id(self):
        self._next_id.incr()
        return self._next_id.get()

    def _get_job_prefix(self):
        return 'J'

    def create_job(self):
        name = '{}{:>05}'.format(self._get_job_prefix(), self._get_next_job_id())
        job = self.add(name)
        return job

    def columns(self):
        return ['Owner', 'Status', 'Created On']

    def _fill_row_cells(self, row, item):
        on = item.created_on.get()
        try:
            on = float(on)
        except:
            pass
        else:
            on = time.ctime(on)

        row.update(dict(
            Owner=item.owner.get(),
            Status=item.status.get(),
        ))
        row['Created On'] = on

    def _fill_row_style(self, style, item, row):
        style['icon'] = item.icon.get()



import os
import gazu

from kabaret import flow
from kabaret.flow_contextual_dict import ContextualView, get_contextual_dict

from libreflow import baseflow
from .kitsu import KitsuEpisode
from .departments import Department
from .file import RenderSettingsTemplate, OutputModuleTemplate


class ShotDepartments(flow.Object):
    layout = flow.Child(Department).ui(expanded=False)
    animation = flow.Child(Department).ui(expanded=False)
    compositing = flow.Child(Department).ui(expanded=False)


class Shot(baseflow.film.Shot):

    _episode = flow.Parent(4)
    departments = flow.Child(ShotDepartments).ui(expanded=True)

    def compute_child_value(self, child_value):
        if child_value is self.kitsu_url:
            child_value.set(
                "%s/%s" % (self._episode.kitsu_url.get(), self.kitsu_id.get())
            )


class Shots(baseflow.film.Shots):

    create_shot = flow.Child(baseflow.maputils.SimpleCreateAction)
    with flow.group("Kitsu"):
        toggle_kitsu_settings = flow.Child(baseflow.film.DisplayKitsuSettings)
        update_kitsu_settings = flow.Child(baseflow.film.UpdateItemsKitsuSettings)

    @classmethod
    def mapped_type(cls):
        return Shot


class Sequence(baseflow.film.Sequence):

    _episode = flow.Parent(2)
    shots = flow.Child(Shots).ui(default_height=420, expanded=True)
    compositing = flow.Child(Department).ui(expanded=True)

    def compute_child_value(self, child_value):
        if child_value is self.kitsu_url:
            child_value.set(
                "%s/shots?search=%s" % (self._episode.kitsu_url.get(), self.name())
            )


class Sequences(baseflow.film.Sequences):

    ICON = ("icons.flow", "sequence")

    _episode = flow.Parent()

    create_sequence = flow.Child(baseflow.maputils.SimpleCreateAction)
    clear_sequences = flow.Child(baseflow.film.ClearSequencesAction).ui(hidden=True)
    update_kitsu_settings = flow.Child(baseflow.film.UpdateItemsKitsuSettings)

    @classmethod
    def mapped_type(cls):
        return Sequence

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return self._episode.get_default_contextual_edits(context_name)


class RenderOperation(flow.values.ChoiceValue):

    CHOICES = ["Rendering", "Preview"]


class RenderAfterEffectsAnimations(flow.Action):

    _episode = flow.Parent()

    operation = flow.Param("Rendering", RenderOperation)

    template_label = flow.Label("<b>Templates:</b>")
    render_settings = flow.Param("majorque_rendu_output", RenderSettingsTemplate)
    output_module = flow.Param("majorque_output", OutputModuleTemplate)

    def run(self, button):
        for seq in self._episode.sequences.mapped_items():
            for shot in seq.shots.mapped_items():
                try:
                    gazu.shot.get_shot(shot.kitsu_id.get())
                except gazu.exception.RouteNotFoundException:
                    pass

                rendering_task_type = gazu.task.get_task_type_by_name("Rendering")
                rendering_task = gazu.task.get_task_by_entity(shot.kitsu_id.get(), rendering_task_type)

                if rendering_task is not None:
                    rendering_task_status = gazu.task.get_task_status(rendering_task["task_status_id"])

                    if rendering_task_status["name"] == "INV":
                        try:
                            anim_file = shot.departments.animation.files["animationA_aep"]
                        except flow.exceptions.MappedNameError:
                            print("No animationA.aep in %s" % shot.departments.animation.oid())
                        else:
                            head_revision = anim_file.get_head_revision()

                            if self.operation.get() == "Rendering":
                                action = head_revision.make_render
                            else:
                                action = head_revision.make_preview
                            
                            action.render_settings.set(self.render_settings.get())
                            action.output_module.set(self.output_module.get())
                            action.run(None)
                            return


class Episode(KitsuEpisode):

    ICON = ("icons.flow", "film")

    settings = flow.Child(ContextualView).ui(hidden=True)
    sequences = flow.Child(Sequences).ui(default_height=420, expanded=True)
    compositing = flow.Child(Department).ui(expanded=False)

    render_aftereffects_animations = flow.Child(RenderAfterEffectsAnimations)

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(episode=self.name())


class Episodes(flow.Map):

    ICON = ("icons.flow", "film")

    create_episode = flow.Child(baseflow.maputils.SimpleCreateAction)

    @classmethod
    def mapped_type(cls):
        return Episode

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(file_category="PROD")



import setuptools
import versioneer
import os

readme = os.path.normpath(os.path.join(__file__, '..', 'README.md'))
with open(readme, "r", encoding="utf-8") as fh:
    long_description = fh.read()

long_description += '\n\n'

changelog = os.path.normpath(os.path.join(__file__, '..', 'CHANGELOG.md'))
with open(changelog, "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    cmdclass=versioneer.get_cmdclass(),
    name="libreflow", # Replace with your own username
    version=versioneer.get_version(),
    author="Flavio Perez",
    author_email="flavio@lfs.coop",
    description="An example flow for kabaret",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/lfs.coop/libreflow",
    license="LGPLv3+",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
    ],
    keywords="kabaret cgwire kitsu gazu animation pipeline libreflow",
    install_requires=["kabaret>=2.2.0rc6",
                      "pyside2~=5.0",
                      "six~=1.0",
                      "kabaret.script-view",
                      "kabaret.subprocess-manager",
                      "kabaret.flow-contextaul-dict",
                      "kabaret.flow-entities",
                      "gazu>=0.8.25",
                      "minio",
                      "timeago",
                      "blender-asset-tracer",
                      "psutil",
                      "fileseq",
                      ],
    python_requires='>=3.7',
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    package_data={
        '': ["*.css", '*.png', '*.svg', '*.gif', '*.abc', '*.aep', '*.ai', '*.blend', '*.jpg', '*.kra', '*.mov', '*.psd', '*.txt', '*.usd', '*.fbx', '*.json', '*.obj', '*.wav'],
    },

)


import pymongo
import re

from kabaret.app._actor import Actor, Cmd, Cmds


#------ CMDS


class SearchCmds(Cmds):
    '''
    '''


@SearchCmds.cmd
class Query_Index(Cmd):
    '''
    Returns a list of dicts describing the first `limit` entries
    of the search index matching the given `query_filter`.

    The dict keys are the following:
        id       : unique id of the entry
        goto_oid : id of the flow object corresponding to this entry
        label    : entry label
    '''
    def _decode(self, query_filter=None, limit=10):
        self._query_filter = query_filter
        self._limit = limit

    def _execute(self):
        results = self.actor().query_index(
            self._query_filter,
            self._limit
        )
        return [r.to_dict() for r in results]


@SearchCmds.cmd
class Query_Project_Index(Cmd):
    '''
    Returns a list of dicts describing the first `limit` entries
    of the search index matching the given `query_filter` and
    referencing objects under the project `project_name`.

    The dict keys are similar to that returned by the Query_Index
    command.
    '''
    def _decode(self, project_name, query_filter=None, limit=10):
        self._project_name = project_name
        self._query_filter = query_filter
        self._limit = limit

    def _execute(self):
        results = self.actor().query_project_index(
            self._project_name,
            self._query_filter,
            self._limit
        )
        return [r.to_dict() for r in results]


@SearchCmds.cmd
class List_Project_Names(Cmd):
    '''
    Returns a list containing the names of all projects
    existing on the cluster.
    '''
    def _decode(self):
        pass

    def _execute(self):
        return self.actor().get_project_names()


#------ Actor


class SearchResult:

    def __init__(self, actor, document):
        self._actor = actor
        self._document = None
        
        self.set_document(document)
    
    def set_document(self, doc):
        self._document = doc
        self._document['id'] = self._document.pop('_id')
    
    def id(self):
        return str(self._document['id'])
    
    def goto_oid(self):
        return self._document['goto_oid']
    
    def label(self):
        return self._document['label']
    
    def to_dict(self):
        return self._document


class Search(Actor):
    '''
    The Search actor manages a single search index for all projects
    defined in the cluster.
    Search entries are indexed in the `Search:index` collection
    within the <CLUSTER_NAME> Mongo database.
    '''

    def __init__(self, session, mongo_uri):
        super(Search, self).__init__(session)
        
        self._mongo_uri = mongo_uri
        self._mongo_client = None
    
    def _create_cmds(self):
        return SearchCmds(self)
    
    def on_session_connected(self):
        self.log('Configuring Search Engine')
        cluster = self.session().get_actor('Cluster')
        cluster_name = cluster.get_cluster_name()
        
        self._mongo_client = pymongo.MongoClient(
            self._mongo_uri,
            appname=self.session().session_uid()
        )
        self._index_collection = self._mongo_client[cluster_name]['Search:index']
    
    def get_result(self, document):
        return SearchResult(self, document)
    
    def query_index(self, query_filter=None, limit=10):
        '''
        Returns the `limit` indexed search entries with the highest
        similarity score with the given filter, ordered by score.
        '''
        # TODO filter using a computed score
        if query_filter is None:
            query_filter = {}
        else:
            tokens = query_filter.split()
            query_filter = {'goto_oid': {'$regex': '.*'.join(tokens)}}
        
        results = [
            self.get_result(doc)
            for doc in self._index_collection.find(query_filter, limit=limit)
        ]
        return results
    
    def get_project_names(self):
        return [
            project_name for project_name, _ in self.session().get_actor('Flow').get_projects_info()
        ]
    
    def query_project_index(self, project_name, query_filter=None, limit=10):
        if query_filter is not None:
            query_filter = f'^/{project_name} {query_filter}'
        
        return self.query_index(query_filter, limit)
    
    def build_project_indexes(self, oid_seed_list=None):
        self._index_collection.delete_many({})
        self.crawl_projects(oid_seed_list)
    
    def crawl_projects(self, oid_seed_list=None):
        '''
        oid_seed_list: list of tuples (oid seed, regex filter)
        '''
        for project_name, _ in self.session().get_actor('Flow').get_projects_info():
            self.crawl_project(project_name, oid_seed_list)
    
    def crawl_project(self, project_name, oid_seed_list=None):
        if oid_seed_list is None:
            # Check if project provides a list of oid seeds
            try:
                oid_seed_list = self.session().cmds.Flow.call(
                    '/'+project_name, 'get_search_oid_seeds', {}, {}
                )
            except AttributeError:
                self.session().log_warning((
                    f'Project {project_name} has not been crawled '
                    'since no oid seed has been provided.'
                ))
                return

        for from_oid, filter, depth in oid_seed_list:
            for oid in self.glob_project(project_name, from_oid, filter, depth):
                self._create_entry(oid)
    
    def glob_project(self, project_name, from_oid, oid_filter, depth):
        if not from_oid.startswith('/'+project_name):
            # Object identified by from_oid doesn't belong to the project
            return []
        
        return self._glob(from_oid, from_oid+'/'+oid_filter, 0, depth)
            
    def _glob(self, from_oid, oid_filter, depth, max_depth=1):
        matches = []

        if re.match(oid_filter, from_oid) is not None:
            matches.append(from_oid)
        
        if depth == max_depth:
            return matches

        for oid in self._ls(from_oid):
            if oid.startswith(from_oid):
                matches += self._glob(oid, oid_filter, depth + 1, max_depth)
        
        return matches
    
    def _ls(self, oid):
        from_relations = self.session().cmds.Flow.ls(oid)
        
        child_oids  = [child_oid for child_oid, _, _, _, _, _ in from_relations[0]]
        child_oids += [oid+'/'+mapped_name for mapped_name in from_relations[1]]
        
        return child_oids
    
    def _create_entry(self, oid):
        self._index_collection.insert_one({
            'goto_oid': oid,
            'label': self.session().cmds.Flow.get_source_display(oid)
        })


import os
import getpass
import re
from collections import defaultdict

from kabaret import flow
from kabaret.flow_contextual_dict import ContextualView, get_contextual_dict

from .file import FileSystemMap
from .users import ToggleBookmarkAction


class CreateDefaultFileSystemItemsAction_XX(flow.Action):

    _department = flow.Parent()

    def allow_context(self, context):
        return context and context.endswith(".inline")

    def _compute_names_and_extensions(self, names_as_string, separator, args):
        formatted_names = names_as_string.split(separator)
        names_and_extensions = []

        for formatted_name in formatted_names:
            formatted_name, extension = tuple(os.path.splitext(formatted_name))
            tracked = "@" not in formatted_name
            formatted_name = formatted_name.replace("@", "")
            name = formatted_name.format_map(defaultdict(str))
            name = re.sub("-", "_", name)
            # name = re.sub('\A_|_\Z', '', name)
            while name.startswith("_"):
                name = name[1:]
            while name.endswith("_"):
                name = name[:-1]

            complete_name = formatted_name.format(**args)
            if extension:
                extension = extension[1:]

            names_and_extensions.append((name, complete_name, extension, tracked))

        return names_and_extensions

    def get_buttons(self):
        self.message.set("<h3>Create default files</h3>")

        settings = get_contextual_dict(self._department, "settings")
        context = settings["context"]
        department = settings["department"]
        default_items = "default_%s_%s_files" % (context, department)

        try:
            default_names_as_string = settings[default_items].replace(" ", "")
        except KeyError:
            msg = self.message.get()
            msg += (
                "<font color=#D5000D>Not default files for %s %s department</font>"
                % (context, department)
            )
            self.message.set(msg)

            return ["Cancel"]

        self._default_names_and_extensions = self._compute_names_and_extensions(
            default_names_as_string, ",", settings
        )

        msg = self.message.get()
        msg += "The following items will be created:<br><br>"

        for (
            name,
            complete_name,
            extension,
            tracked,
        ) in self._default_names_and_extensions:
            msg += "  %s%s" % ("" if tracked else "@", name)
            if extension:
                msg += ".%s" % extension
            msg += " (%s)<br>" % complete_name

        self.message.set(msg)

        return ["Confirm", "Cancel"]

    def run(self, button):
        if button == "Cancel":
            return

        settings = get_contextual_dict(self._department, "settings")
        settings = defaultdict(str, settings)

        for (
            name,
            complete_name,
            extension,
            tracked,
        ) in self._default_names_and_extensions:
            try:
                if extension:  # File
                    if tracked:
                        self._department.files.add_tracked_file(
                            name, extension, complete_name
                        )
                    else:
                        self._department.files.add_file(name, extension)
                else:  # Folder
                    if tracked:
                        self._department.files.add_tracked_folder(name, complete_name)
                    else:
                        self._department.files.add_folder(name)
            except ValueError:
                # Item has already been created manually
                pass

        self._department.files.touch()


class CreateDefaultFileSystemItemsAction(flow.Action):

    _department = flow.Parent()

    def _ensure_file(self, name, extension):
        files = self._department.files

        if not files.has_mapped_name('%s_%s' % (name, extension)):
            files.create_file.file_name.set(name)
            files.create_file.file_format.set(extension)
            files.create_file.run(None)
    
    def _ensure_folder(self, name):
        files = self._department.files

        if not files.has_mapped_name(name):
            files.create_folder.folder_name.set(name)
            files.create_folder.run(None)
    
    def needs_dialog(self):
        return False
    
    def allow_context(self, context):
        return context and context.endswith('.inline')

    def run(self, button):
        if button == 'Cancel':
            return

        settings = get_contextual_dict(self._department, 'settings')
        default_files = self.root().project().admin.project_settings.default_files

        names = default_files['%s_%s' % (settings['context'], settings['department'])].file_names.get()
        names = names.replace(' ', '').split(',')

        for filename in names:
            name, ext = os.path.splitext(filename)

            if ext:
                self._ensure_file(name, ext[1:])
            else:
                self._ensure_folder(name)
        
        self._department.files.touch()


class ContextualEnv(flow.Object):

    department = flow.Parent(2)
    environment = flow.Child(ContextualView)


class GotoContextualEnv(flow.Action):

    _department = flow.Parent()

    def needs_dialog(self):
        return False
    
    def allow_context(self, context):
        return context and context.endswith('.inline')
    
    def run(self, button):
        oid = self._department.contextual_env.oid()
        return self.get_result(goto=oid)


class Department(flow.Object):

    _parent = flow.Parent()

    toggle_bookmark = flow.Child(ToggleBookmarkAction)

    settings = flow.Child(ContextualView).ui(hidden=True)
    contextual_env = flow.Child(ContextualEnv).ui(hidden=True)
    files = (
        flow.Child(FileSystemMap)
        .injectable()
        .ui(
            expanded=True,
            # auto_fit=False,
            # columns_width=(10, 40, 50),
            action_submenus=True,
            items_action_submenus=True
        )
    )
    auto_current = flow.BoolParam(True).ui(hidden=True)
    show_environment = flow.Child(GotoContextualEnv)
    path = flow.Computed(cached=True).ui(hidden=True)

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(
                department=self.name(),
                context=self._parent.__class__.__name__.lower(),
            )

    def get_contextual_view(self, context_name):
        if context_name == 'environment':
            return self.contextual_env.environment
    
    def compute_child_value(self, child_value):
        if child_value is self.path:
            settings = get_contextual_dict(self, "settings")
            file_category = settings["file_category"]
            if file_category == "PROD":
                path = os.path.join(
                    file_category,
                    settings["sequence"],
                    settings["shot"],
                    settings["department"],
                )
            elif file_category == "LIB":
                path = os.path.join(
                    file_category,
                    settings["asset_name"],
                    settings["department"],
                )

            child_value.set(path)
